
/* Chrysalide - Outil d'analyse de fichiers binaires
 * processor.c - équivalent Python du fichier "arch/processor.c"
 *
 * Copyright (C) 2015 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  OpenIDA is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  OpenIDA is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "processor.h"


#include <pygobject.h>


#include <i18n.h>


#include <arch/processor.h>


#include "instruction.h"
#include "vmpa.h"
#include "../helpers.h"




















/* ------------------ MANIPULATIONS DES INSTRUCTIONS DESASSEMBLEES ------------------ */



/* Fournit les instructions désassemblées pour une architecture. */
static PyObject *py_arch_processor_get_disass_instrs(PyObject *, void *);

/* Recherche une instruction d'après son adresse. */
static PyObject *py_arch_processor_find_instr_by_addr(PyObject *, PyObject *);

/* Fournit l'instruction qui en précède une autre. */
static PyObject *py_arch_processor_get_prev_instr(PyObject *, PyObject *);

/* Fournit l'instruction qui en suit une autre. */
static PyObject *py_arch_processor_get_next_instr(PyObject *, PyObject *);





/* ---------------------------------------------------------------------------------- */
/*                    MANIPULATIONS DES INSTRUCTIONS DESASSEMBLEES                    */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Fournit les instructions désassemblées pour une architecture.*
*                                                                             *
*  Retour      : Liste des instructions désassemblées ou None si aucune.      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_arch_processor_get_disass_instrs(PyObject *self, void *closure)
{
    PyObject *result;                       /* Instance Python à retourner */
    GArchProcessor *proc;                   /* Architecture visée          */
    GArchInstruction *instrs;               /* Série d'instructions liées  */

    proc = G_ARCH_PROCESSOR(pygobject_get(self));
    instrs = g_arch_processor_get_disassembled_instructions(proc);

    if (instrs != NULL)
        result = pygobject_new(G_OBJECT(instrs));
    else
    {
        result = Py_None;
        Py_INCREF(result);
    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                value   = valeur fournie à intégrer ou prendre en compte.    *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Note les instructions désassemblées avec une architecture.   *
*                                                                             *
*  Retour      : Bilan de l'opération pour Python.                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static int py_arch_processor_set_disass_instrs(PyObject *self, PyObject *value, void *closure)
{
    GArchProcessor *proc;                   /* Architecture visée          */
    GArchInstruction *instrs;               /* Série d'instructions liées  */

    if (!PyObject_TypeCheck(value, get_python_arch_instruction_type()))
    {
        PyErr_SetString(PyExc_TypeError, _("The attribute value must be an instruction."));
        return -1;
    }

    proc = G_ARCH_PROCESSOR(pygobject_get(self));
    instrs = G_ARCH_INSTRUCTION(pygobject_get(value));

    g_arch_processor_set_disassembled_instructions(proc, instrs);

    return 0;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = processeur d'architecture à manipuler.                *
*                args = instruction représentant le point de départ.          *
*                                                                             *
*  Description : Recherche une instruction d'après son adresse.               *
*                                                                             *
*  Retour      : Instruction trouvée à l'adresse donnée, None si aucune.      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_arch_processor_find_instr_by_addr(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Instance à retourner        */
    PyObject *addr_obj;                     /* Objet pour une localisation */
    int ret;                                /* Bilan de lecture des args.  */
    GArchProcessor *proc;                   /* Processeur manipulé         */
    vmpa2t *addr;                           /* Localisation à retrouver    */
    GArchInstruction *found;                /* Instruction liée trouvée    */

    ret = PyArg_ParseTuple(args, "O", &addr_obj);
    if (!ret) return NULL;

    ret = PyObject_IsInstance(addr_obj, (PyObject *)get_python_vmpa_type());
    if (!ret) return NULL;

    proc = G_ARCH_PROCESSOR(pygobject_get(self));
    addr = get_internal_vmpa(addr_obj);

    found = g_arch_processor_find_instr_by_address(proc, addr);

    if (found != NULL)
        result = pygobject_new(G_OBJECT(found));

    else
    {
        result = Py_None;
        Py_INCREF(result);
    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = processeur d'architecture à manipuler.                *
*                args = instruction représentant le point de départ.          *
*                                                                             *
*  Description : Fournit l'instruction qui en précède une autre.              *
*                                                                             *
*  Retour      : Instruction précédente trouvée, ou NULL.                     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_arch_processor_get_prev_instr(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Instance à retourner        */
    PyObject *instr_obj;                    /* Objet pour une instruction  */
    int ret;                                /* Bilan de lecture des args.  */
    GArchProcessor *proc;                   /* Processeur manipulé         */
    GArchInstruction *instr;                /* Instruction de référence    */
    GArchInstruction *found;                /* Instruction liée trouvée    */

    ret = PyArg_ParseTuple(args, "O", &instr_obj);
    if (!ret) return NULL;

    ret = PyObject_IsInstance(instr_obj, (PyObject *)get_python_arch_instruction_type());
    if (!ret) return NULL;

    proc = G_ARCH_PROCESSOR(pygobject_get(self));
    instr = G_ARCH_INSTRUCTION(pygobject_get(instr_obj));

    found = g_arch_processor_get_prev_instr(proc, instr);

    if (found != NULL)
        result = pygobject_new(G_OBJECT(found));

    else
    {
        result = Py_None;
        Py_INCREF(result);
    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = processeur d'architecture à manipuler.                *
*                args = instruction représentant le point de départ.          *
*                                                                             *
*  Description : Fournit l'instruction qui en suit une autre.                 *
*                                                                             *
*  Retour      : Instruction suivante trouvée, ou NULL.                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_arch_processor_get_next_instr(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Instance à retourner        */
    PyObject *instr_obj;                    /* Objet pour une instruction  */
    int ret;                                /* Bilan de lecture des args.  */
    GArchProcessor *proc;                   /* Processeur manipulé         */
    GArchInstruction *instr;                /* Instruction de référence    */
    GArchInstruction *found;                /* Instruction liée trouvée    */

    ret = PyArg_ParseTuple(args, "O", &instr_obj);
    if (!ret) return NULL;

    ret = PyObject_IsInstance(instr_obj, (PyObject *)get_python_arch_instruction_type());
    if (!ret) return NULL;

    proc = G_ARCH_PROCESSOR(pygobject_get(self));
    instr = G_ARCH_INSTRUCTION(pygobject_get(instr_obj));

    found = g_arch_processor_get_next_instr(proc, instr);

    if (found != NULL)
        result = pygobject_new(G_OBJECT(found));

    else
    {
        result = Py_None;
        Py_INCREF(result);
    }

    return result;

}


















/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit un accès à une définition de type à diffuser.        *
*                                                                             *
*  Retour      : Définition d'objet pour Python.                              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

PyTypeObject *get_python_arch_processor_type(void)
{
    static PyMethodDef py_arch_processor_methods[] = {
        {
            "find_instr_by_addr", py_arch_processor_find_instr_by_addr,
            METH_VARARGS,
            "find_instr_by_addr($self, addr, /)\n--\n\nLook for an instruction located at a given address."
        },
        {
            "get_prev_instr", py_arch_processor_get_prev_instr,
            METH_VARARGS,
            "get_prev_instr($self, instr, /)\n--\n\nProvide the instruction preceding a given instruction."
        },
        {
            "get_next_instr", py_arch_processor_get_next_instr,
            METH_VARARGS,
            "get_next_instr($self, instr, /)\n--\n\nProvide the instruction following a given instruction."
        },
        { NULL }
    };

    static PyGetSetDef py_arch_processor_getseters[] = {
        {
            "disass_instrs", py_arch_processor_get_disass_instrs, py_arch_processor_set_disass_instrs,
            "Give access to the disassembled instructions run by the current processor.", NULL
        },
        { NULL }
    };

    static PyTypeObject py_arch_processor_type = {

        PyVarObject_HEAD_INIT(NULL, 0)

        .tp_name        = "pychrysalide.arch.ArchProcessor",
        .tp_basicsize   = sizeof(PyGObject),

        .tp_flags       = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_BASETYPE,

        .tp_doc         = "PyChrysalide processor for a given architecture.",

        .tp_methods     = py_arch_processor_methods,
        .tp_getset      = py_arch_processor_getseters,

    };

    return &py_arch_processor_type;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : module = module dont la définition est à compléter.          *
*                                                                             *
*  Description : Prend en charge l'objet 'pychrysalide.arch.ArchProcessor'.   *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool register_python_arch_processor(PyObject *module)
{
    PyTypeObject *py_arch_processor_type;   /* Type Python 'BinContent'    */
    int ret;                                /* Bilan d'un appel            */
    PyObject *dict;                         /* Dictionnaire du module      */

    py_arch_processor_type = get_python_arch_processor_type();

    py_arch_processor_type->tp_base = &PyGObject_Type;
    py_arch_processor_type->tp_basicsize = py_arch_processor_type->tp_base->tp_basicsize;

    APPLY_ABSTRACT_FLAG(py_arch_processor_type);

    if (PyType_Ready(py_arch_processor_type) != 0)
        return false;

    Py_INCREF(py_arch_processor_type);
    ret = PyModule_AddObject(module, "ArchProcessor", (PyObject *)py_arch_processor_type);
    if (ret != 0) return false;

    dict = PyModule_GetDict(module);
    pygobject_register_class(dict, "ArchProcessor", G_TYPE_ARCH_PROCESSOR, py_arch_processor_type,
                             Py_BuildValue("(O)", py_arch_processor_type->tp_base));

    return true;

}
